ROOT="/"
FIRMWARE_INSTALL_PATH="#{ROOT}/protonet/firmware/"
CONFIG_BASE="#{ROOT}/data/var/lib/protonet/config/"

require 'digest/md5'
require 'fileutils'
require 'open-uri'
require 'json'
require File.dirname(__FILE__) + '/protonet_config'
require File.dirname(__FILE__) + '/version'
require File.dirname(__FILE__) + '/firmware_logger'

class FirmwareUpdater
  def initialize(version)
    @version = version
  end

  def exists?(md5, path)
    (Digest::MD5.file(path).hexdigest == md5) rescue false
  end

  def remove_tempfiles(filename)
    # We don't care about possible FileUtils exceptions for the temporary path file
    `rm #{filename}_path #{filename} >& /dev/null`
  end

  def download_and_check(md5, path)
    retry_count = 0
    puts  "downloading #{path}"
    FileUtils.mkdir_p(File.dirname(path))

    temp_file = '.download_temp'
    temp_path_file = temp_file + '_path'
    loop do
      puts  "retrying #{path}" if retry_count >= 1

      # make sure that we're trying to resume the right file
      # this can happen when the firmware changes between retries
      # good luck using unless for this simple condition
      if !File.exists?(temp_path_file) || (File.exists?(temp_path_file) && File.read(temp_path_file) != path)
        remove_tempfiles(temp_file)
        File.open(temp_path_file, 'w') {|f| f.write(path) }
      end
      dl_status = system "curl -s -C - -o #{temp_file} #{FIRMWARE_BASE_URL}/#{path}  && mv #{temp_file} #{path}"

      FileUtils.rm temp_path_file if exists?(md5, path)

      break if retry_count >= 3 || exists?(md5, path)

      # download was successful, but cksum is invalid, delete and try again
      remove_tempfiles(temp_file) if dl_status && !exists?(md5, path)

      retry_count += 1
    end

    # throw exception if 3 retries are not enough
    raise unless exists?(md5, path)
  end


  def load_manifest
    image_files = open("#{FIRMWARE_BASE_URL}/#{@version}/MANIFEST").
      read.split("\n").
      map { |line| line.split(/\s+/) }
  end

  def ensure_downloaded
    puts  "updating #{@version}"

    firmware_files = load_manifest

    firmware_files.each do |md5, path|
      next if exists?(md5, path)
      download_and_check(md5, path)
    end

    files_not_in_firmware = Dir["#{@version}/**/**"].reject { |path| File.directory?(path) }.sort - firmware_files.map { |i| i[1] }.sort

    if files_not_in_firmware.count > 0
      FileUtils.rm(files_not_in_firmware)      
      puts  "deleted #{files_not_in_firmware.join(", ")}"
    end

    FileUtils.ln(IMAGE, "#{@version}/image")

    true # all good!
  end

  def activate_kernel
    puts  "activating #{KERNEL}"

    begin
      `dpkg --force-all -i #{KERNEL}` 
      $?.exitstatus == 0
    rescue Exception => e
      puts "Cannot install kernel! #{e.inspect}"
      false
    end
  end
end

def generate_protonet_config
  config = protonet_config
  puts  "generating config"
  File.open("#{CONFIG_BASE}/current", "w") { |f| f.write JSON.pretty_generate(config) }
  FileUtils.chmod 0664, "#{CONFIG_BASE}/current"
  FileUtils.chown 'protonet', 'protonet', "#{CONFIG_BASE}/current"
  true
rescue Exception => e
  puts "ERROR: cannot generate config! #{e.inspect}"
  false
end


def prep_system
  FileUtils.mkdir_p(FIRMWARE_INSTALL_PATH)
  FileUtils.mkdir_p(CONFIG_BASE)

  Dir.chdir(ROOT) do 
    FileUtils.rm_f("firmware")
    FileUtils.ln_s("./protonet/firmware", "firmware")
  end

  true
end

def activate_firmware(version)
  Dir.chdir(FIRMWARE_INSTALL_PATH) do 
    puts "activating #{version}"

    FileUtils.rm_f('current')
    FileUtils.ln_s("./#{version}", "current")
  end
end

def download_and_activate_firmware(version)
  update_firmware(version) && activate_firmware(version)
  true
end

def update_firmware(version)
  Dir.chdir(FIRMWARE_INSTALL_PATH) do 
    updater = FirmwareUpdater.new(version)
    updater.ensure_downloaded && updater.activate_kernel
  end
end

def migrate_protonet_data
  return unless File.directory?('/home/protonet/dashboard')
  puts "migrate_protonet_data"

  `sv down /home/protonet/dashboard/shared/services/enabled/*`

  `mkdir -p /data/home/protonet/dashboard/shared`
  `mv /home/protonet/dashboard/shared/files /data/home/protonet/dashboard/shared`
  `mv /home/protonet/dashboard/shared/public /data/home/protonet/dashboard/shared`
  `mv /home/protonet/dashboard/shared/log /data/home/protonet/dashboard/shared`
  `mkdir -p /data/var/lib`
  `mv /var/lib/mysql /data/var/lib`
  `mv /var/lib/libvirt /data/var/lib`
  `mkdir -p /data/etc`
  `mv /etc/libvirt /data/etc`
  `mv /home/protonet/dashboard /home/protonet/inactive_dashboard`
  `mv /etc/sv/protonet_runit /etc/protonet_runit.inactive`
end

def undo_migrate_protonet_data
  return unless File.directory?('/home/protonet/inactive_dashboard')
  puts "undo_migrate_protonet_data"

  `mv /data/home/protonet/dashboard/shared/files /home/protonet/dashboard/shared/`
  `mv /data/home/protonet/dashboard/shared/public /home/protonet/dashboard/shared/`
  `mv /data/home/protonet/dashboard/shared/log /home/protonet/dashboard/shared/`
  `mv /data/var/lib/libvirt /var/lib`
  `mv /data/etc/libvirt /etc`
  `mv /home/protonet/inactive_dashboard /home/protonet/dashboard`
  `mv /etc/protonet_runit.inactive /etc/sv/protonet_runit`
end

def with_lockfile(lockfile)
  file = File.open(lockfile, 'w+')
  if file.flock(File::LOCK_NB | File::LOCK_EX)
    yield
  else
    puts "cannot aquire #{lockfile}"
  end
ensure
  file.close
end

version = PROTONET_VERSION
prep_system

with_lockfile(FIRMWARE_INSTALL_PATH+'firmware_updater.lock') do
  if ARGV[0] == "update"
    update_firmware(version) && exit(0)
  elsif ARGV[0] == "install"
    if File.exists?('/protonet/firmware/current')
      download_and_activate_firmware(version) && exit(0)
    else
      # upgrade legacy system
      generate_protonet_config && download_and_activate_firmware(version) && migrate_protonet_data && exit(0)
    end
  end
end

exit(1)
